/*    Copyright 2012 10gen Inc.
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

#include "mongo/platform/basic.h"

#include <iostream>

#include <boost/static_assert.hpp>
#include <climits>

#include "counter.h"
#include "mongo/unittest/unittest.h"

namespace mongo {
    namespace {
        TEST( CounterTest, Test1 ) {
            Counter64 c;
            ASSERT_EQUALS(c.get(), 0);
            c.increment();
            ASSERT_EQUALS(c.get(), 1);
            c.decrement();
            ASSERT_EQUALS(c.get(), 0);
            c.decrement(3);
            ASSERT_EQUALS(c.get(), -3);
            c.increment(1);
            ASSERT_EQUALS(c.get(), -2);
            c.decrement(-1);
            ASSERT_EQUALS(c.get(), -1);
            c.increment();
            ASSERT_EQUALS(static_cast<long long>(c), 0);
        }

    }  // namespace
}  // namespace mongo
